/*
 * Unisphere -- UNIformly distributing points on SPHERE
 * 
 * Copyright (C) 2010 by Ali Katanforoush <a_katanforosh@sbu.ac.ir>
 *
 * This file is a part of Unisphere.
 *
 * Unisphere is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Unisphere is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
// SPoptim.cpp
// =============================================================================
//
// Console Application: take input data on command line arguments.
//
// =============================================================================

#include "unisphere.h"
#include "uncoptim.h"

#include <string.h>
#include <stdio.h>

static char optName[7][5]={"dfp", "bfgs", "pr", "prr", "fr", "frr", "init"};
static char lsName[4][13]={"GOLDENSEARCH", "BSSEARCH", "SSEARCH", "RSSEARCH"};

class uniS2:public UnconstrainedOptim {
public:
		int np;
		int metric;	// 0: log(r), 1: 1/r, 2: 1/r^2
		int optim;	// 0: DFP,  1: BFGS,  2: PR,  3: PRR,  4: FR,  5:FRR

    uniS2(int NP, int METRIC, int OPTIM=3, int lsmethod=RSSEARCH)
            :UnconstrainedOptim(2*NP,0,0,lsmethod) {
        np=NP;
        metric=METRIC;
        optim=OPTIM;
    }
    void readInit(char *filename);
    void reInit();
    void randInit();
    void icolatInit(int ml, int nl);
    void outFile(char *filename);
    void outFile();
    void optimize();

    void F() {
        f=0;
        int i,j;
        switch(metric){
        case 0:
            for(i=0; i<np; ++i) {
                for(j=i+1; j<np; ++j) {
                    double dt=x[2*i]-x[2*j];
                    double df=x[2*i+1]-x[2*j+1];
                    double sf=x[2*i+1]+x[2*j+1];
                    double a=cos(sf);
                    double b=cos(df);
                    double c=cos(dt);
                    double t=2+(1-c)*a-(1+c)*b;
                    f-=log(t/2);
                }
            }
						break;
        case 1:
            for(i=0; i<np; ++i) {
                for(j=i+1; j<np; ++j) {
                    double dt=x[2*i]-x[2*j];
                    double df=x[2*i+1]-x[2*j+1];
                    double sf=x[2*i+1]+x[2*j+1];
                    double a=cos(sf);
                    double b=cos(df);
                    double c=cos(dt);
                    double t=2+(1-c)*a-(1+c)*b;
                    f+=1/t;
                }
            }
						break;
        case 2:
            for(i=0; i<np; ++i) {
                for(j=i+1; j<np; ++j) {
                    double dt=x[2*i]-x[2*j];
                    double df=x[2*i+1]-x[2*j+1];
                    double sf=x[2*i+1]+x[2*j+1];
                    double a=cos(sf);
                    double b=cos(df);
                    double c=cos(dt);
                    double t=2+(1-c)*a-(1+c)*b;
                    t*=t;
                    f+=1/t;
                }
            }
				}
    }

    void G() {
        int i,j;
        for(i=0; i<np; ++i) {
            g[2*i]=g[2*i+1]=0;
        }
        switch(metric){
        case 0:
            for(i=0; i<np; ++i) {
                for(j=i+1; j<np; ++j) {
                    double dt=x[2*i]-x[2*j];
                    double df=x[2*i+1]-x[2*j+1];
                    double sf=x[2*i+1]+x[2*j+1];
                    double a=cos(sf);
                    double b=cos(df);
                    double c=cos(dt);
                    double t=2+(1-c)*a-(1+c)*b;
                    double gt=sin(dt)*(a+b)/t;
                    g[2*i]-=gt;
                    g[2*j]+=gt;
                    double gt1=(1+c)*sin(df);
                    double gt2=(1-c)*sin(sf);
                    g[2*i+1]+=(gt2-gt1)/t;
                    g[2*j+1]+=(gt2+gt1)/t;
                }
            }
						break;
        case 1:
            for(i=0; i<np; ++i) {
                for(j=i+1; j<np; ++j) {
                    double dt=x[2*i]-x[2*j];
                    double df=x[2*i+1]-x[2*j+1];
                    double sf=x[2*i+1]+x[2*j+1];
                    double a=cos(sf);
                    double b=cos(df);
                    double c=cos(dt);
                    double t=2+(1-c)*a-(1+c)*b;
                    double t2=t*t;
                    double gt=sin(dt)*(a+b)/t2;
                    g[2*i]-=gt;
                    g[2*j]+=gt;
                    double gt1=(1+c)*sin(df);
                    double gt2=(1-c)*sin(sf);
                    g[2*i+1]+=(gt2-gt1)/t2;
                    g[2*j+1]+=(gt2+gt1)/t2;
                }
            }
						break;
        case 2:
            for(i=0; i<np; ++i) {
                for(j=i+1; j<np; ++j) {
                    double dt=x[2*i]-x[2*j];
                    double df=x[2*i+1]-x[2*j+1];
                    double sf=x[2*i+1]+x[2*j+1];
                    double a=cos(sf);
                    double b=cos(df);
                    double c=cos(dt);
                    double t=2+(1-c)*a-(1+c)*b;
                    double t3=t*t*t/2;
                    double gt=sin(dt)*(a+b)/t3;
                    g[2*i]-=gt;
                    g[2*j]+=gt;
                    double gt1=(1+c)*sin(df);
                    double gt2=(1-c)*sin(sf);
                    g[2*i+1]+=(gt2-gt1)/t3;
                    g[2*j+1]+=(gt2+gt1)/t3;
                }
            }
        }
    }
};

void uniS2::readInit(char *filename) {
    FILE *fp;
    fp=fopen(filename,"r");
    double *p=x;
    for(int i=0; i<np; ++i) {
        fscanf(fp, "%lf %lf\n", p, p+1);
        ++p; ++p;
    }
    fclose(fp);
}

void uniS2::reInit() {
    int N;
    double e;
    np=interpolateRE(np,N,e);

    double *fi=new double[np];

    eqdistre(N,e,x,fi);

    for(int i=np-1; i>=0; --i) {
        x[2*i]=x[i];
        x[2*i+1]=fi[i];
    }
    delete fi;
}

void uniS2::icolatInit(int ml, int nl) {
    double *y=new double[np];
    double *z=new double[np];
    cs3icolattice(ml, nl, x, y, z);

    int i;
    for(i=0; i<np; ++i) {
        y[i]=arg(x[i],y[i]);
        z[i]=asin(z[i]);
    }
    for(i=np-1; i>=0; --i) {
        x[2*i]=y[i];
        x[2*i+1]=z[i];
    }
    delete y;
    delete z;
}

void uniS2::randInit() {
    double *fi=new double[np];
    ps3rand(np,x,fi);
    for(int i=np-1; i>=0; --i) {
        x[2*i]=x[i];
        x[2*i+1]=fi[i];
    }
    delete fi;
}

void uniS2::outFile(char *filename) {
    FILE *fp;
    fp=fopen(filename, "w");
    double *p=x;
    int i=np;
    while(--i>=0) {
        fprintf(fp, "%30.25lg %30.25lg\n", *p, *(p+1));
        ++p; ++p;
        if(i%100==0)
            fflush(fp);
    }
    fclose(fp);
}

void uniS2::outFile() {
    char filename[80];
    filename[0]='\0';
    switch(metric){
    case 0:
        sprintf(filename, ".\\S2L-%0i.opt.%s", np, optName[optim]);
        break;
    case 1:
        sprintf(filename, ".\\S2H-%0i.opt.%s", np, optName[optim]);
        break;
    case 2:
        sprintf(filename, ".\\S2G-%0i.opt.%s", np, optName[optim]);
		}
    outFile(filename);
}

void uniS2::optimize() {
    switch(optim) {
    case 0:
        DFPoptimize();
				break;
    case 1:
        BFGSoptimize();
				break;
    case 2:
        PRoptimize(0);
				break;
    case 3:
        PRoptimize(1);
				break;
    case 4:
        FRoptimize(0);
				break;
    case 5:
				FRoptimize(1);
				break;
		default:
				F();
		}
}


int main(int argc, char *argv[]) {
		if(argc<=1 || argv[1][0]=='h' || argv[1][1]=='h'){
			fprintf(stderr, "Syntax:\n");
			fprintf(stderr, "   SPoptime Np metric [optimization_method] [line_search_method] [inital_configuration] [outfile]\n\n");
			fprintf(stderr, "Np : Number of Points\n");
			fprintf(stderr, "metric : G (gravity potential, 1/r^2) | H (coulomb potential, 1/r) | L (weakest potential, 1/log(r)\n");
			fprintf(stderr, "optimization_method : dfp | bfgs | pr | prr (default) | fr | frr | init (without optimization)\n");
			fprintf(stderr, "line_search_method : GOLDENSEARCH | BSSEARCH | SSEARCH | RSSEARCH (default)\n");
			fprintf(stderr, "inital_configuration : R (random) | P (polar coordinates) | L <m> <n> (lattice points) | <input_file> \n");
			fprintf(stderr, "outfile : <output_file> \n");
			fprintf(stderr, "\n\n Notice that in lattice points method you must determine two numbers m and n which satisfy Np=10*(m*m+m*n+n*n)+2\n");

			return 1;
		}

		int lastarg=1;

		int NP=912;
		if(argc>1) {
				NP=atoi(argv[1]);
				++lastarg;
		}

		int METRIC=0;
		if(argc>2) {
				switch(argv[2][0]) {
				case 'G':
						METRIC=2;
						++lastarg;
						break;
				case 'H':
						METRIC=1;
						++lastarg;
						break;
				case 'L':
						METRIC=0;
						++lastarg;
				}
		}

		int OPTIM=3;
		if(argc>3) {
				for(OPTIM=0; OPTIM<7 && strcmp(optName[OPTIM], argv[3])!=0; ++OPTIM)
						;
				if(OPTIM==7)
						OPTIM=3;
				else
						++lastarg;
		}

		int lsmethod=RSSEARCH;
		if(argc>4) {
				for(lsmethod=0; lsmethod<4 && strcmp(lsName[lsmethod], argv[4])!=0; ++lsmethod)
						;
				if(lsmethod==4)
						lsmethod=RSSEARCH;
				else
						++lastarg;
		}

		uniS2 s(NP, METRIC, OPTIM, lsmethod);

		if(argc>5) {
				switch(argv[5][0]) {
				case 'R':
						s.randInit();
						++lastarg;
						break;
				case 'P':
						s.reInit();
						++lastarg;
						break;
				case 'L':
						s.icolatInit(atoi(argv[6]), atoi(argv[7]));
						++lastarg;
						++lastarg;
						++lastarg;
						break;
				default:
						s.readInit(argv[5]);
						++lastarg;
				}
		}
		else
			s.reInit();


		s.optimize();

		if(argc>lastarg)
				s.outFile(argv[lastarg]);
		else
				s.outFile();

		printf("Total iteration: %i   f evaluation: %i   g evaluation: %i\n", s.cnls, s.cnf, s.cng);
		printf(" Object: %50.30lg\n", s.f);

		return 1;
}

